"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getInstance = exports.getConnection = void 0;

var _neo4jDriver = _interopRequireDefault(require("neo4j-driver"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

/**
 * Singleton connection driver Neo4j
 *
 * @return {Neo4jDriver}
 */
var instance = null;

var getConnection = function getConnection() {
  // its already instanced
  if (instance) {
    return instance;
  } // get info from env


  var _getEnv = getEnv(),
      connection_string = _getEnv.connection_string,
      NEO4J_USERNAME = _getEnv.NEO4J_USERNAME,
      NEO4J_PASSWORD = _getEnv.NEO4J_PASSWORD,
      config = _getEnv.config; // create instance


  instance = _neo4jDriver["default"].driver(connection_string, _neo4jDriver["default"].auth.basic(NEO4J_USERNAME, NEO4J_PASSWORD), config);
  return instance;
};
/**
 *
 * Generate Object config using .env configuration
 *
 * @return {Object}
 */


exports.getConnection = getConnection;

var getEnv = function getEnv() {
  var _process$env = process.env,
      _process$env$NEO4J_PR = _process$env.NEO4J_PROTOCOL,
      NEO4J_PROTOCOL = _process$env$NEO4J_PR === void 0 ? 'bolt' : _process$env$NEO4J_PR,
      _process$env$NEO4J_HO = _process$env.NEO4J_HOST,
      NEO4J_HOST = _process$env$NEO4J_HO === void 0 ? 'localhost' : _process$env$NEO4J_HO,
      _process$env$NEO4J_PO = _process$env.NEO4J_PORT,
      NEO4J_PORT = _process$env$NEO4J_PO === void 0 ? '7687' : _process$env$NEO4J_PO,
      _process$env$NEO4J_US = _process$env.NEO4J_USERNAME,
      NEO4J_USERNAME = _process$env$NEO4J_US === void 0 ? 'neo4j' : _process$env$NEO4J_US,
      _process$env$NEO4J_PA = _process$env.NEO4J_PASSWORD,
      NEO4J_PASSWORD = _process$env$NEO4J_PA === void 0 ? 'letmein' : _process$env$NEO4J_PA,
      NEO4J_ENTERPRISE = _process$env.NEO4J_ENTERPRISE,
      _process$env$NEO4J_DA = _process$env.NEO4J_DATABASE,
      NEO4J_DATABASE = _process$env$NEO4J_DA === void 0 ? 'neo4j' : _process$env$NEO4J_DA;
  var connection_string = "".concat(NEO4J_PROTOCOL, "://").concat(NEO4J_HOST, ":").concat(NEO4J_PORT);
  var enterprise = NEO4J_ENTERPRISE === 'true'; // Build additional config

  var config = {};
  var settings = {
    NEO4J_ENCRYPTION: 'encrypted',
    NEO4J_TRUST: 'trust',
    NEO4J_TRUSTED_CERTIFICATES: 'trustedCertificates',
    NEO4J_KNOWN_HOSTS: 'knownHosts',
    NEO4J_MAX_CONNECTION_POOLSIZE: 'maxConnectionPoolSize',
    NEO4J_MAX_TRANSACTION_RETRY_TIME: 'maxTransactionRetryTime',
    NEO4J_LOAD_BALANCING_STRATEGY: 'loadBalancingStrategy',
    NEO4J_MAX_CONNECTION_LIFETIME: 'maxConnectionLifetime',
    NEO4J_CONNECTION_TIMEOUT: 'connectionTimeout',
    NEO4J_DISABLE_LOSSLESS_INTEGERS: 'disableLosslessIntegers',
    NEO4J_LOGGING_LEVEL: 'logging'
  };
  Object.keys(settings).forEach(function (setting) {
    if (process.env.hasOwnProperty(setting)) {
      var key = settings[setting];
      var value = process.env[setting];

      if (key == 'trustedCertificates') {
        value = value.split(',');
      } else if (key == 'disableLosslessIntegers') {
        value = value === 'true';
      }

      config[key] = value;
    }
  });
  return {
    connection_string: connection_string,
    NEO4J_USERNAME: NEO4J_USERNAME,
    NEO4J_PASSWORD: NEO4J_PASSWORD,
    enterprise: enterprise,
    NEO4J_DATABASE: NEO4J_DATABASE,
    config: config
  };
};

var getInstance = function getInstance() {
  return _neo4jDriver["default"];
};

exports.getInstance = getInstance;