"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Model = void 0;

var _cypher = require("./cypher");

var _collection = require("./collection");

var _utils = require("./utils");

var _hydrate = require("./hydrate");

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(n); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _iterableToArrayLimit(arr, i) { if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return; var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var ORDER_BY_FUNCTIONS_ALLOWED = ['toUpper', 'toLower', 'left', 'lTrim', 'replace', 'right', 'rTrim', 'trim', 'substring', 'toString'];

var Model = /*#__PURE__*/function () {
  /**
   * Constructor
   *
   * @param {Object} values
   */
  function Model() {
    var _this = this;

    var values = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var labels = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var attributes = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _classCallCheck(this, Model);

    this._with = [];
    this._values = values;
    this._labels = labels;
    this._attributes = attributes;
    this._alias = null;
    this.filter_attributes = [];
    this.errors = {};
    Object.entries(attributes).forEach(function (_ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          key = _ref2[0],
          field = _ref2[1];

      (0, _utils.createGetterAndSetter)(_this, key, field.set, field.get);
    });
    Object.entries(values).forEach(function (_ref3) {
      var _ref4 = _slicedToArray(_ref3, 2),
          key = _ref4[0],
          value = _ref4[1];

      _this[key] = value;
    });
  }
  /**
   * Start the retrieve Info based on actual Node
   */


  _createClass(Model, [{
    key: "toJSON",
    value: function toJSON() {
      return this.retriveInfo(this);
    }
    /**
     * Retrieve Info from Node as a JSON, only with clean data
     *
     * @param {Object} model
     */

  }, {
    key: "retriveInfo",
    value: function retriveInfo(model, previous) {
      var _this2 = this;

      var data = {};
      data.id = model.id; //attributes of relations

      if (previous) {
        for (var _i2 = 0, _Object$entries = Object.entries(previous.attributes); _i2 < _Object$entries.length; _i2++) {
          var _Object$entries$_i = _slicedToArray(_Object$entries[_i2], 1),
              relKey = _Object$entries$_i[0];

          if (model._values[relKey]) data[relKey] = model._values[relKey];
        }
      }

      Object.entries(model._attributes).forEach(function (_ref5) {
        var _ref6 = _slicedToArray(_ref5, 2),
            key = _ref6[0],
            field = _ref6[1];

        switch (field.type) {
          case 'hash':
            break;

          case 'relationship':
            if (model._values[key]) data[key] = _this2.retriveInfo(model._values[key], field);
            break;

          case 'relationships':
            if (model._values[key]) {
              data[key] = Object.values(model._values[key]).map(function (item) {
                return _this2.retriveInfo(item, field);
              });
            }

            break;

          default:
            data[key] = model[key];
        }
      });
      return data;
    }
  }, {
    key: "getAliasName",
    value: function getAliasName() {
      var _this$_alias;

      return (_this$_alias = this._alias) !== null && _this$_alias !== void 0 ? _this$_alias : this._labels.join('').toLowerCase();
    }
  }, {
    key: "getNodeName",
    value: function getNodeName() {
      return this._labels.join(':');
    }
  }, {
    key: "getCypherName",
    value: function getCypherName() {
      var aliasName = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      if (aliasName) {
        return aliasName + ':' + this.getNodeName();
      }

      return this.getAliasName() + ':' + this.getNodeName();
    }
  }, {
    key: "getAttributes",
    value: function getAttributes() {
      return Object.entries(this._attributes);
    }
  }, {
    key: "writeFilter",
    value: function writeFilter(forNode) {
      var _this3 = this;

      var relationAlias = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
      // FILTERS WITH LOCATION
      this.filter_attributes.filter(function (item) {
        return item["for"] === forNode || item["for"] === relationAlias;
      }).forEach(function (_ref7) {
        var attr = _ref7.attr,
            operator = _ref7.operator,
            value = _ref7.value;

        _this3.cypher.addWhere({
          attr: attr,
          operator: operator,
          value: value
        });
      });
      this.cypher.matchs.push(this.cypher.writeWhere());
    }
  }, {
    key: "writeOrderBy",
    value: function writeOrderBy() {
      var _this4 = this;

      // FILTERS WITH LOCATION
      this.order_by.forEach(function (_ref8) {
        var attr = _ref8.attr,
            direction = _ref8.direction;

        _this4.cypher.addOrderBy(attr, direction);
      });
    }
  }, {
    key: "doMatchs",
    value: function doMatchs(node, relation) {
      var _relation$previousNod,
          _this5 = this;

      var level = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;

      if (relation) {
        this.cypher.match(relation.previousNode, relation.previousAlias, relation.relationship, node);
        this.cypher.modelReturn(relation.previousAlias, node);
      } else {
        this.cypher.match(node);
        this.cypher.modelReturn(node.getAliasName(), node);
      }

      this.writeFilter(node.getAliasName(), "".concat(relation === null || relation === void 0 ? void 0 : (_relation$previousNod = relation.previousNode) === null || _relation$previousNod === void 0 ? void 0 : _relation$previousNod.getAliasName(), "_").concat(relation === null || relation === void 0 ? void 0 : relation.previousAlias));
      Object.keys(node._attributes).forEach(function (key) {
        var field = node._attributes[key];

        if (field.isModel) {
          var _checkWith = (0, _hydrate.checkWith)(level, key, _this5._with),
              _checkWith2 = _slicedToArray(_checkWith, 2),
              found_condition = _checkWith2[0],
              isOptional = _checkWith2[1];

          if (found_condition) {
            var newNode = new field.target(undefined, node._state);

            _this5.cypher.modelReturnRelation("".concat(node.getAliasName(), "_").concat(key), field);

            newNode.filter_relationship = field.filter_relationship;
            newNode._alias = key;
            newNode.isOptional = isOptional;
            newNode.collectFirst = !field.isArray;

            _this5.doMatchs(newNode, {
              relationship: ":".concat(field.labels.join(':')),
              previousNode: node,
              previousAlias: key
            }, level + 1);
          }
        }
      });
      return true;
    }
  }, {
    key: "addMatchs",
    value: function addMatchs(node, field) {
      this.cypher.match(node, false, false, false, field.attr);
      this.cypher.modelReturn(field.attr, node);
      this.cypher.modelReturnRelation(field.relationName, field);
      this.writeFilter(field.attr, "".concat(node.getAliasName(), "_").concat(field.attr));
    }
  }, {
    key: "fetch",
    value: function () {
      var _fetch = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee() {
        var with_related,
            state,
            _args = arguments;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                with_related = _args.length > 0 && _args[0] !== undefined ? _args[0] : [];
                state = _args.length > 1 && _args[1] !== undefined ? _args[1] : undefined;
                // reset alias to default
                this._alias = this._labels.join('').toLowerCase(); // return a hydrated findAll

                return _context.abrupt("return", this.constructor.findAll({
                  filter_attributes: [{
                    key: "id(".concat(this.getAliasName(), ")"),
                    value: this.id
                  }],
                  with_related: with_related,
                  state: state,
                  parent: this
                }));

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function fetch() {
        return _fetch.apply(this, arguments);
      }

      return fetch;
    }()
  }, {
    key: "delete",
    value: function () {
      var _delete2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2() {
        var _this6 = this;

        var detach,
            _args2 = arguments;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                detach = _args2.length > 0 && _args2[0] !== undefined ? _args2[0] : false;
                this.cypher = new _cypher.Cypher();
                this.filter_attributes = [{
                  key: "id(".concat(this.getAliasName(), ")"),
                  value: this.id
                }].map(function (fa) {
                  return _this6.prepareFilter(fa, _this6);
                });
                this.doMatchs(this, false);
                return _context2.abrupt("return", this.cypher["delete"](this.getAliasName(), detach));

              case 5:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function _delete() {
        return _delete2.apply(this, arguments);
      }

      return _delete;
    }()
  }, {
    key: "setAttributes",
    value: function setAttributes() {
      var _this7 = this;

      var create = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : true;
      this.errors = {};
      Object.entries(this._attributes).forEach(function (_ref9) {
        var _ref10 = _slicedToArray(_ref9, 2),
            key = _ref10[0],
            field = _ref10[1];

        _this7._values[key] = field.getDefaultValue(_this7._values[key]);

        try {
          field.checkValidation(key, _this7._values[key]);

          if (field.isModel === false) {
            _this7.cypher.addSet(_this7.getAliasName() + '.' + key, _this7._values[key]);
          } else if (create) {// TODO: add the relation
          }
        } catch (e) {
          var error = JSON.parse(e.message);
          _this7.errors[error.key] = error.msg;
        }
      });
      if (Object.keys(this.errors).length > 0) throw new Error('Model invalid, check the .errors attribute');
    }
  }, {
    key: "isValid",
    value: function isValid() {
      var _this8 = this;

      var ret = true;
      this.errors = {};
      Object.entries(this._attributes).forEach(function (_ref11) {
        var _ref12 = _slicedToArray(_ref11, 2),
            key = _ref12[0],
            field = _ref12[1];

        try {
          field.checkValidation(key, field.getDefaultValue(_this8._values[key]));
        } catch (e) {
          var error = JSON.parse(e.message);
          _this8.errors[error.key] = error.msg;
          ret = false;
        }
      });
      return ret;
    }
  }, {
    key: "save",
    value: function () {
      var _save = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3() {
        var record, _record;

        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                this.cypher = new _cypher.Cypher();

                if (!(this.id === undefined)) {
                  _context3.next = 10;
                  break;
                }

                // create
                this.doMatchs(this, false);
                this.setAttributes(false);
                _context3.next = 6;
                return this.cypher.create(this.getCypherName());

              case 6:
                record = _context3.sent;
                (0, _hydrate.hydrate)(this, record, this._with);
                _context3.next = 18;
                break;

              case 10:
                // update
                this.cypher.addWhere({
                  attr: "id(".concat(this.getAliasName(), ")"),
                  value: this.id
                });
                this.cypher.isDistinct();
                this.doMatchs(this, false);
                this.setAttributes();
                _context3.next = 16;
                return this.cypher.update();

              case 16:
                _record = _context3.sent;
                (0, _hydrate.hydrate)(this, _record[0], this._with);

              case 18:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function save() {
        return _save.apply(this, arguments);
      }

      return save;
    }()
    /**
     * Relate nodes
     *
     * @param {String} attr
     * @param {Model} node
     * @param {JSON} attributes
     */

  }, {
    key: "relate",
    value: function () {
      var _relate = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(attr, node) {
        var _this9 = this;

        var attributes,
            create,
            field,
            _i3,
            _Object$entries2,
            _Object$entries2$_i,
            relKey,
            relField,
            value,
            error,
            data,
            _args4 = arguments;

        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                attributes = _args4.length > 2 && _args4[2] !== undefined ? _args4[2] : {};
                create = _args4.length > 3 && _args4[3] !== undefined ? _args4[3] : true;
                // ADD TO _WITH TO RETURN THE RELATION
                this._with = [];
                this.cypher = new _cypher.Cypher();
                this.filter_attributes = [{
                  key: "id(".concat(this.getAliasName(), ")"),
                  value: this.id
                }, {
                  key: "id(".concat(attr, ")"),
                  value: node.id
                }].map(function (fa) {
                  return _this9.prepareFilter(fa, _this9);
                });
                this.doMatchs(this); // CREATE THE RELATION FOR THIS ATTR

                field = this._attributes[attr];

                if (field) {
                  _context4.next = 9;
                  break;
                }

                throw new Error("Attribute \"".concat(attr, "\" does not exists on model \"").concat(this.getAliasName(), "\""));

              case 9:
                field.attr = attr;
                field.relationName = "".concat(this.getAliasName(), "_").concat(attr);
                this.addMatchs(node, field); // ADD TO _WITH TO RETURN THE RELATION

                this._with = [[attr]]; // ADD THE ATTRIBUTES ON RELATION

                if (field.attributes) {
                  this.errors = {};

                  for (_i3 = 0, _Object$entries2 = Object.entries(field.attributes); _i3 < _Object$entries2.length; _i3++) {
                    _Object$entries2$_i = _slicedToArray(_Object$entries2[_i3], 2), relKey = _Object$entries2$_i[0], relField = _Object$entries2$_i[1];
                    value = relField.getDefaultValue(attributes[relKey]);

                    try {
                      relField.checkValidation(relKey, value);
                    } catch (e) {
                      error = JSON.parse(e.message);
                      this.errors[error.key] = error.msg;
                    }

                    if (value !== undefined) this.cypher.addSet(this.getAliasName() + '_' + attr + '.' + relKey, relField.set(value));
                  }
                }

                if (!(Object.keys(this.errors).length > 0)) {
                  _context4.next = 16;
                  break;
                }

                throw new Error('Relationship invalid, check the .errors attribute');

              case 16:
                _context4.next = 18;
                return this.cypher.relate(this, field, node, create);

              case 18:
                data = _context4.sent;
                data.forEach(function (record) {
                  (0, _hydrate.hydrate)(_this9, record, _this9._with);
                });

              case 20:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function relate(_x, _x2) {
        return _relate.apply(this, arguments);
      }

      return relate;
    }()
    /**
     * Update a relation between the nodes
     *
     * @param {String} attr
     * @param {Model} node
     * @param {JSON} attributes
     */

  }, {
    key: "updateRelationship",
    value: function () {
      var _updateRelationship = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(attr, node) {
        var attributes,
            _args5 = arguments;
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                attributes = _args5.length > 2 && _args5[2] !== undefined ? _args5[2] : {};
                return _context5.abrupt("return", this.relate(attr, node, attributes, false));

              case 2:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function updateRelationship(_x3, _x4) {
        return _updateRelationship.apply(this, arguments);
      }

      return updateRelationship;
    }()
    /**
     * Create a relation between the nodes
     *
     * @param {String} attr
     * @param {Model} node
     * @param {JSON} attributes
     */

  }, {
    key: "createRelationship",
    value: function () {
      var _createRelationship = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(attr, node) {
        var attributes,
            _args6 = arguments;
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                attributes = _args6.length > 2 && _args6[2] !== undefined ? _args6[2] : {};
                return _context6.abrupt("return", this.relate(attr, node, attributes, true));

              case 2:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function createRelationship(_x5, _x6) {
        return _createRelationship.apply(this, arguments);
      }

      return createRelationship;
    }()
    /**
     * Remove the relations about that attribute
     *
     * @param {String} attr
     */

  }, {
    key: "removeAllRelationships",
    value: function () {
      var _removeAllRelationships = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(attr) {
        var _this10 = this;

        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                this.cypher = new _cypher.Cypher();
                this._with = [[attr]];
                this.filter_attributes = [{
                  key: "id(".concat(this.getAliasName(), ")"),
                  value: this.id
                }].map(function (fa) {
                  return _this10.prepareFilter(fa, _this10);
                });
                this.doMatchs(this);
                return _context7.abrupt("return", this.cypher["delete"]("".concat(this.getAliasName(), "_").concat(attr)));

              case 5:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function removeAllRelationships(_x7) {
        return _removeAllRelationships.apply(this, arguments);
      }

      return removeAllRelationships;
    }()
    /**
     * Remove the one single relationship based on other node
     *
     * @param {String} attr
     */

  }, {
    key: "removeRelationship",
    value: function () {
      var _removeRelationship = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8(attr, node) {
        var _this11 = this;

        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                this.cypher = new _cypher.Cypher();
                this._with = [[attr]];
                this.filter_attributes = [{
                  key: "id(".concat(this.getAliasName(), ")"),
                  value: this.id
                }, {
                  key: "id(".concat(attr, ")"),
                  value: node.id
                }].map(function (fa) {
                  return _this11.prepareFilter(fa, _this11);
                });
                this.doMatchs(this);
                return _context8.abrupt("return", this.cypher["delete"]("".concat(this.getAliasName(), "_").concat(attr)));

              case 5:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function removeRelationship(_x8, _x9) {
        return _removeRelationship.apply(this, arguments);
      }

      return removeRelationship;
    }()
    /**
     * Create a new relation and remove the older
     *
     * @param {String} attr
     * @param {Model} node
     * @param {JSON} attributes
     */

  }, {
    key: "recreateRelationship",
    value: function () {
      var _recreateRelationship = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee9(attr, node) {
        var attributes,
            data,
            _args9 = arguments;
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                attributes = _args9.length > 2 && _args9[2] !== undefined ? _args9[2] : {};
                _context9.prev = 1;
                _context9.next = 4;
                return this.removeAllRelationships(attr);

              case 4:
                _context9.next = 8;
                break;

              case 6:
                _context9.prev = 6;
                _context9.t0 = _context9["catch"](1);

              case 8:
                _context9.prev = 8;
                _context9.next = 11;
                return this.relate(attr, node, attributes, true);

              case 11:
                data = _context9.sent;
                return _context9.abrupt("return", data);

              case 15:
                _context9.prev = 15;
                _context9.t1 = _context9["catch"](8);
                throw new Error('new relation is not possible');

              case 18:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, this, [[1, 6], [8, 15]]);
      }));

      function recreateRelationship(_x10, _x11) {
        return _recreateRelationship.apply(this, arguments);
      }

      return recreateRelationship;
    }()
  }, {
    key: "prepareFilter",
    value: function prepareFilter(fa, model) {
      if (!fa) return false;
      var isCypherFunction = /.+\(.+\)/.test(fa.key);

      if (isCypherFunction) {
        var regExp = /\(([^)]+)\)/;
        var matches = regExp.exec(fa.key); //matches[1] contains the value between the parentheses

        fa["for"] = matches[1];
        fa.attr = fa.key;
      } else {
        fa["for"] = fa.key.split('.').length > 1 ? fa.key.split('.')[0] : model.getAliasName();
        fa.attr = fa.key.split('.').length > 1 ? fa.key : "".concat(model.getAliasName(), ".").concat(fa.key);
      }

      return fa;
    }
  }], [{
    key: "findByID",
    value: function () {
      var _findByID = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee10(id) {
        var config,
            self,
            data,
            _args10 = arguments;
        return regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                config = _args10.length > 1 && _args10[1] !== undefined ? _args10[1] : {};
                self = new this();
                config.filter_attributes = [{
                  key: "id(".concat(self.getAliasName(), ")"),
                  value: parseInt(id, 10)
                }].concat(config.filter_attributes);
                _context10.next = 5;
                return this.findAll(config);

              case 5:
                data = _context10.sent;
                return _context10.abrupt("return", data.first());

              case 7:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function findByID(_x12) {
        return _findByID.apply(this, arguments);
      }

      return findByID;
    }()
  }, {
    key: "findBy",
    value: function () {
      var _findBy = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee11() {
        var filter_attributes,
            config,
            _args11 = arguments;
        return regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                filter_attributes = _args11.length > 0 && _args11[0] !== undefined ? _args11[0] : [];
                config = _args11.length > 1 && _args11[1] !== undefined ? _args11[1] : {};
                config.filter_attributes = filter_attributes;
                return _context11.abrupt("return", this.findAll(config));

              case 4:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11, this);
      }));

      function findBy() {
        return _findBy.apply(this, arguments);
      }

      return findBy;
    }()
  }, {
    key: "findAll",
    value: function () {
      var _findAll = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee12() {
        var _this12 = this;

        var config,
            self,
            data,
            result,
            ids,
            _args12 = arguments;
        return regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                config = _args12.length > 0 && _args12[0] !== undefined ? _args12[0] : {};

                if (!config.parent) {
                  self = new this(undefined, config.state);
                  self._state = config.state;
                } else {
                  self = config.parent;
                  self.parent = true;
                }

                Object.keys(config).forEach(function (key) {
                  config[key] === undefined && delete config[key];
                });
                config = Object.assign({
                  with_related: [],
                  filter_attributes: [],
                  onlyRelation: false,
                  order_by: [],
                  skip: '',
                  limit: '',
                  optional: true,
                  state: undefined
                }, config);
                config.with_related.forEach(function (item) {
                  var w = item.split('__');

                  self._with.push(w);
                }); // setWith(self._with)

                self.cypher = new _cypher.Cypher(); // self.cypher.isDistinct()

                self.cypher.optional = config.optional;
                self.cypher.skip = config.skip;
                self.cypher.limit = config.limit;
                self.filter_attributes = config.filter_attributes.map(function (fa) {
                  return self.prepareFilter(fa, self);
                });
                self.order_by = config.order_by.map(function (ob) {
                  var isCypherFunction = /.+\(.+\)/.test(ob.key);

                  if (isCypherFunction) {
                    var regExp = /(.+)\(([^)]+)\)/;
                    var matches = regExp.exec(ob.key);
                    if (!ORDER_BY_FUNCTIONS_ALLOWED.includes(matches[1])) throw new Error("Function (".concat(matches[1], ") are not allowed in order_by"));
                    ob["for"] = matches[2];
                    ob.attr = ob.key;
                  } else {
                    ob["for"] = ob.key.split('.').length > 1 ? ob.key.split('.')[0] : self.getAliasName();
                    ob.attr = ob.key.split('.').length > 1 ? ob.key : "".concat(self.getAliasName(), ".").concat(ob.key);
                  }

                  return ob;
                });
                self.doMatchs(self, false, 0);
                self.writeOrderBy();
                _context12.next = 15;
                return self.cypher.find();

              case 15:
                data = _context12.sent;
                result = new _collection.Collection();
                ids = [];
                data.forEach(function (record) {
                  var model = new _this12(undefined, config.state);
                  model._state = config.state;

                  var main = record._fields[record._fieldLookup[model.getAliasName()]];

                  var id = (0, _utils.convertID)(main.id);

                  if (config.parent) {
                    model = config.parent;
                  } else {
                    if (ids.includes(id)) {
                      model = result[ids.indexOf(id)];
                    } else {
                      ids.push(id);
                    }
                  }

                  result[ids.indexOf(id)] = (0, _hydrate.hydrate)(model, record, self._with);
                });
                return _context12.abrupt("return", result);

              case 20:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12, this);
      }));

      function findAll() {
        return _findAll.apply(this, arguments);
      }

      return findAll;
    }()
  }]);

  return Model;
}();

exports.Model = Model;